#include "RakNetTransport.h"
#include "RakNetworkFactory.h"
#include "RakPeerInterface.h"
#include "BitStream.h"
#include "PacketEnumerations.h"
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#if defined(__GNUC__)
#define _vsnprintf vsnprintf
#endif

RakNetTransportCommandParser::RakNetTransportCommandParser()
{
	RegisterCommand(1, "SetPassword","Changes the console password to whatever.");
	RegisterCommand(0, "ClearPassword","Removes the console passwords.");
	RegisterCommand(0, "GetPassword","Gets the console password.");
}
RakNetTransportCommandParser::~RakNetTransportCommandParser()
{
}
bool RakNetTransportCommandParser::OnCommand(const char *command, unsigned numParameters, char **parameterList, TransportInterface *transport, PlayerID playerId, const char *originalString)
{
	RakNetTransport *rnt = (RakNetTransport*) transport;
	if (strcmp(command, "SetPassword")==0)
	{
		rnt->SetIncomingPassword(parameterList[0]);
		rnt->Send(playerId, "Password changed to %s\r\n", parameterList[0]);
	}
	else if (strcmp(command, "ClearPassword")==0)
	{
		rnt->SetIncomingPassword(0);
		rnt->Send(playerId, "Password cleared\r\n");
	}
	else if (strcmp(command, "GetPassword")==0)
	{
		unsigned bytesUsed;
		char text[REMOTE_MAX_TEXT_INPUT];
		memset(text, 0, sizeof(text));
		RakNet::BitStream *bs;
		bs=rnt->GetIncomingPassword();
		bytesUsed=bs->GetNumberOfBytesUsed();
		if (bytesUsed >= REMOTE_MAX_TEXT_INPUT )
			bytesUsed=REMOTE_MAX_TEXT_INPUT-1;
		memcpy(text, bs->GetData(), bytesUsed);
		if (text[0])
			rnt->Send(playerId, "Password is %s\r\n",text);
		else
			rnt->Send(playerId, "No password is set.\r\n");
	}
	return true;
}
char *RakNetTransportCommandParser::GetName(void) const
{
	return "RakNetTransport";
}
void RakNetTransportCommandParser::SendHelp(TransportInterface *transport, PlayerID playerId)
{
	transport->Send(playerId, "Provides a secure connection between your console\r\n");
	transport->Send(playerId, "and the console server.  Used to modify the console password.\r\n");
}
RakNetTransport::RakNetTransport()
{
	rakPeer=0;
}
RakNetTransport::~RakNetTransport()
{
	if (rakPeer)
		RakNetworkFactory::DestroyRakPeerInterface(rakPeer);
}
bool RakNetTransport::Start(unsigned short port, bool serverMode)
{
	AutoAllocate();
	rakPeer->InitializeSecurity(0,0,0,0);

	if (serverMode)
	{
		// Allow up to 8 remote systems to login
		rakPeer->SetMaximumIncomingConnections(8);
	}

	return rakPeer->Initialize(8, port, 250, 0);
}
void RakNetTransport::Stop(void)
{
	if (rakPeer==0) return;
	rakPeer->Disconnect(1000, 0);
	newConnections.clear();
	lostConnections.clear();
}
void RakNetTransport::Send( PlayerID playerId, const char *data, ... )
{
	if (rakPeer==0) return;
	if (data==0 || data[0]==0) return;

	char text[REMOTE_MAX_TEXT_INPUT];
	va_list ap;
	va_start(ap, data);
	_vsnprintf(text, REMOTE_MAX_TEXT_INPUT, data, ap);
	va_end(ap);
	text[REMOTE_MAX_TEXT_INPUT-1]=0;

	RakNet::BitStream str;
	str.Write((unsigned char) ID_TRANSPORT_STRING);
	str.Write(text, (int) strlen(text));
	str.Write((unsigned char) 0); // Null terminate the string
	rakPeer->Send(&str, MEDIUM_PRIORITY, RELIABLE_ORDERED, 0, playerId, (bool) (playerId==UNASSIGNED_PLAYER_ID));
}
void RakNetTransport::CloseConnection( PlayerID playerId )
{
	rakPeer->CloseConnection(playerId, true, 0);
}
Packet* RakNetTransport::Receive( void )
{
	if (rakPeer==0) return 0;
	Packet *p;
	p=rakPeer->Receive();
	if (p==0)
		return 0;
	if (p->data[0]==ID_TRANSPORT_STRING)
	{
		p->data++; // Go past ID_TRANSPORT_STRING, since the transport protocol is only supposed to send strings.
		return p;
	}
	if (p->data[0]==ID_NEW_INCOMING_CONNECTION)
	{
		newConnections.push(p->playerId);
	}
	else if (p->data[0]==ID_DISCONNECTION_NOTIFICATION || p->data[0]==ID_CONNECTION_LOST)
	{
		lostConnections.push(p->playerId);
	}
	rakPeer->DeallocatePacket(p);

	return 0;
}
PlayerID RakNetTransport::HasNewConnection(void)
{
	if (newConnections.size())
		return newConnections.pop();
	return UNASSIGNED_PLAYER_ID;
}
PlayerID RakNetTransport::HasLostConnection(void)
{
	if (lostConnections.size())
		return lostConnections.pop();
	return UNASSIGNED_PLAYER_ID;
}
void RakNetTransport::SetIncomingPassword(const char *password)
{
	if (password)
		rakPeer->SetIncomingPassword(password, (int) strlen(password));
	else
		rakPeer->SetIncomingPassword(0, 0);
}
RakNet::BitStream * RakNetTransport::GetIncomingPassword(void)
{
	return rakPeer->GetIncomingPassword();
}
void RakNetTransport::DeallocatePacket( Packet *packet )
{
	if (rakPeer==0) return;
	packet->data--; // Go back to ID_TRANSPORT_STRING, which we passed up in Receive()
	rakPeer->DeallocatePacket(packet);
}
void RakNetTransport::AutoAllocate(void)
{
	if (rakPeer==0)
		rakPeer=RakNetworkFactory::GetRakPeerInterface();
}
CommandParserInterface* RakNetTransport::GetCommandParser(void)
{
    return &rakNetTransportCommandParser;
}
